@@ -47,81 +47,61 @@ void buffer_free(buffer *b) {
 	free(b);
 }
 
-void buffer_reset(buffer *b) {
+__attribute_cold__
+static void buffer_free_ptr(buffer *b) {
 	if (NULL == b) return;
-
-	/* limit don't reuse buffer larger than ... bytes */
-	if (b->size > BUFFER_MAX_REUSE_SIZE) {
-		free(b->ptr);
-		b->ptr = NULL;
-		b->size = 0;
-	} else if (b->size > 0) {
-		b->ptr[0] = '\0';
-	}
-
+	free(b->ptr);
+	b->ptr = NULL;
 	b->used = 0;
+	b->size = 0;
+}
+
+void buffer_reset(buffer *b) {
+	if (NULL != b && b->size > 0) {
+		b->used = 0;
+		/* release buffer larger than ... bytes */
+		if (b->size > BUFFER_MAX_REUSE_SIZE) buffer_free_ptr(b);
+	}
 }
 
 void buffer_move(buffer *b, buffer *src) {
 	buffer tmp;
+	force_assert(NULL != b);
+	force_assert(NULL != src);
 
-	if (NULL == b) {
-		buffer_reset(src);
-		return;
-	}
-	buffer_reset(b);
-	if (NULL == src) return;
-
+	buffer_clear(b);
 	tmp = *src; *src = *b; *b = tmp;
 }
 
-#define BUFFER_PIECE_SIZE 64
-static size_t buffer_align_size(size_t size) {
-	size_t align = BUFFER_PIECE_SIZE - (size % BUFFER_PIECE_SIZE);
-	/* overflow on unsinged size_t is defined to wrap around */
-	if (size + align < size) return size;
-	return size + align;
+/* make sure buffer is at least "size" big + 1 for '\0'. keep old data */
+__attribute_cold__
+static void buffer_realloc(buffer *b, size_t len) {
+    #define BUFFER_PIECE_SIZE 64uL  /*(must be power-of-2)*/
+    const size_t sz = (len + 1 + BUFFER_PIECE_SIZE-1) & ~(BUFFER_PIECE_SIZE-1);
+    force_assert(sz > len);
+
+    b->size = sz;
+    b->ptr = realloc(b->ptr, sz);
+
+    force_assert(NULL != b->ptr);
 }
 
-/* make sure buffer is at least "size" big. discard old data */
-static void buffer_alloc(buffer *b, size_t size) {
-	force_assert(NULL != b);
-	if (0 == size) size = 1;
-
-	if (size <= b->size) return;
-
-	if (NULL != b->ptr) free(b->ptr);
-
-	b->used = 0;
-	b->size = buffer_align_size(size);
-	b->ptr = malloc(b->size);
-
-	force_assert(NULL != b->ptr);
+__attribute_cold__
+static void buffer_alloc_replace(buffer *b, size_t size) {
+    /*(discard old data so realloc() does not copy)*/
+    if (NULL != b->ptr) {
+        free(b->ptr);
+        b->ptr = NULL;
+    }
+    buffer_realloc(b, size);
 }
 
-/* make sure buffer is at least "size" big. keep old data */
-static void buffer_realloc(buffer *b, size_t size) {
-	force_assert(NULL != b);
-	if (0 == size) size = 1;
-
-	if (size <= b->size) return;
-
-	b->size = buffer_align_size(size);
-	b->ptr = realloc(b->ptr, b->size);
-
-	force_assert(NULL != b->ptr);
-}
-
-
 char* buffer_string_prepare_copy(buffer *b, size_t size) {
 	force_assert(NULL != b);
-	force_assert(size + 1 > size);
 
-	buffer_alloc(b, size + 1);
-
-	b->used = 1;
-	b->ptr[0] = '\0';
+	if (size >= b->size) buffer_alloc_replace(b, size);
 
+	b->used = 0;
 	return b->ptr;
 }
 
@@ -149,10 +129,7 @@ char* buffer_string_prepare_append(buffer *b, size_t size) {
 void buffer_string_set_length(buffer *b, size_t len) {
 	force_assert(NULL != b);
 
-	if (len >= b->size) {
-		force_assert(len + 1 > len);
-		buffer_realloc(b, len + 1);
-	}
+	if (len >= b->size) buffer_realloc(b, len);
 
 	b->used = len + 1;
 	b->ptr[len] = '\0';
@@ -167,10 +144,10 @@ void buffer_commit(buffer *b, size_t size)
 
 	if (size > 0) {
 		/* check for overflow: unsigned overflow is defined to wrap around */
-		force_assert(b->used + size > b->used);
-
-		force_assert(b->used + size <= b->size);
-		b->used += size;
+		size_t sz = b->used + size;
+		force_assert(sz > b->used);
+		force_assert(sz <= b->size);
+		b->used = sz;
 	}
 
 	b->ptr[b->used - 1] = '\0';
@@ -184,22 +161,13 @@ void buffer_copy_string_len(buffer *b, const char *s, size_t s_len) {
 	force_assert(NULL != b);
 	force_assert(NULL != s || s_len == 0);
 
-	buffer_string_prepare_copy(b, s_len);
+	if (s_len >= b->size) buffer_string_prepare_copy(b, s_len);
 
 	if (0 != s_len) memcpy(b->ptr, s, s_len); /*(s might be NULL)*/
 	b->ptr[s_len] = '\0';
 	b->used = s_len + 1;
 }
 
-void buffer_copy_buffer(buffer *b, const buffer *src) {
-	if (NULL == src || 0 == src->used) {
-		buffer_string_prepare_copy(b, 0);
-		b->used = 0; /* keep special empty state for now */
-	} else {
-		buffer_copy_string_len(b, src->ptr, buffer_string_length(src));
-	}
-}
-
 void buffer_append_string(buffer *b, const char *s) {
 	buffer_append_string_len(b, s, NULL != s ? strlen(s) : 0);
 }
@@ -222,20 +190,27 @@ void buffer_append_string_len(buffer *b, const char *s, size_t s_len) {
 	force_assert(NULL != s || s_len == 0);
 
 	target_buf = buffer_string_prepare_append(b, s_len);
+	if (0 == b->used) ++b->used; /*(must include '\0' for append below)*/
 
-	if (0 == s_len) return; /* nothing to append *//*(s might be NULL)*/
-
-	memcpy(target_buf, s, s_len);
+	/*(s might be NULL if 0 == s_len)*/
+	if (s_len) memcpy(target_buf, s, s_len);
 	target_buf[s_len] = '\0';
 	b->used += s_len;
 }
 
-void buffer_append_string_buffer(buffer *b, const buffer *src) {
-	if (NULL == src) {
-		buffer_append_string_len(b, NULL, 0);
-	} else {
-		buffer_append_string_len(b, src->ptr, buffer_string_length(src));
-	}
+void buffer_append_path_len(buffer *b, const char *a, size_t alen) {
+    size_t blen = buffer_string_length(b);
+    int aslash = (alen && a[0] == '/');
+    buffer_string_prepare_append(b, alen+2); /*(+ '/' and + '\0' if 0 == blen)*/
+    if (blen && b->ptr[blen-1] == '/') {
+        if (aslash) --b->used;
+    }
+    else {
+        if (!b->used) ++b->used;
+        if (!aslash) b->ptr[++b->used - 2] = '/';
+    }
+    memcpy(b->ptr+b->used-1, a, alen);
+    b->ptr[(b->used += alen)-1] = '\0';
 }
 
 void buffer_append_uint_hex_lc(buffer *b, uintmax_t value) {
@@ -306,14 +281,9 @@ void buffer_append_strftime(buffer *b, const char *format, const struct tm *tm)
 	size_t r;
 	char* buf;
 	force_assert(NULL != b);
+	force_assert(NULL != format);
 	force_assert(NULL != tm);
 
-	if (NULL == format || '\0' == format[0]) {
-		/* empty format */
-		buffer_string_prepare_append(b, 0);
-		return;
-	}
-
 	buf = buffer_string_prepare_append(b, 255);
 	r = strftime(buf, buffer_string_space(b), format, tm);
 
@@ -729,21 +699,18 @@ void buffer_append_string_c_escaped(buffer *b, const char *s, size_t s_len) {
 
 
 void buffer_copy_string_encoded_cgi_varnames(buffer *b, const char *s, size_t s_len, int is_http_header) {
-	size_t i, j;
+	size_t i, j = 0;
 
 	force_assert(NULL != b);
 	force_assert(NULL != s || 0 == s_len);
 
-	buffer_reset(b);
+	buffer_string_prepare_copy(b, s_len + 5);
 
 	if (is_http_header && NULL != s && 0 != strcasecmp(s, "CONTENT-TYPE")) {
-		buffer_string_prepare_append(b, s_len + 5);
 		buffer_copy_string_len(b, CONST_STR_LEN("HTTP_"));
-	} else {
-		buffer_string_prepare_append(b, s_len);
+		j = 5; /* "HTTP_" */
 	}
 
-	j = buffer_string_length(b);
 	for (i = 0; i < s_len; ++i) {
 		unsigned char cr = s[i];
 		if (light_isalpha(cr)) {
@@ -820,6 +787,39 @@ void buffer_urldecode_query(buffer *url) {
 	buffer_urldecode_internal(url, 1);
 }
 
+int buffer_is_valid_UTF8(const buffer *b) {
+    /* https://www.w3.org/International/questions/qa-forms-utf-8 */
+    const unsigned char *c = (unsigned char *)b->ptr;
+    while (*c) {
+
+        /*(note: includes ctrls)*/
+        if (                         c[0] <  0x80 ) { ++c;  continue; }
+
+        if (         0xc2 <= c[0] && c[0] <= 0xdf
+            &&       0x80 <= c[1] && c[1] <= 0xbf ) { c+=2; continue; }
+
+        if ( (   (   0xe0 == c[0]
+                  && 0xa0 <= c[1] && c[1] <= 0xbf)
+              || (   0xe1 <= c[0] && c[0] <= 0xef && c[0] != 0xed
+                  && 0x80 <= c[1] && c[1] <= 0xbf)
+              || (   0xed == c[0]
+                  && 0x80 <= c[1] && c[1] <= 0x9f)   )
+            &&       0x80 <= c[2] && c[2] <= 0xbf ) { c+=3; continue; }
+
+        if ( (   (   0xf0 == c[0]
+                  && 0x90 <= c[1] && c[1] <= 0xbf)
+              || (   0xf1 <= c[0] && c[0] <= 0xf3
+                  && 0x80 <= c[1] && c[1] <= 0xbf)
+              || (   0xf4 == c[0]
+                  && 0x80 <= c[1] && c[1] <= 0x8f)   )
+            &&       0x80 <= c[2] && c[2] <= 0xbf
+            &&       0x80 <= c[3] && c[3] <= 0xbf ) { c+=4; continue; }
+
+        return 0; /* invalid */
+    }
+    return 1; /* valid */
+}
+
 /* - special case: empty string returns empty string
  * - on windows or cygwin: replace \ with /
  * - strip leading spaces
@@ -848,19 +848,12 @@ void buffer_path_simplify(buffer *dest, buffer *src)
 	force_assert(NULL != dest && NULL != src);
 
 	if (buffer_string_is_empty(src)) {
-		buffer_string_prepare_copy(dest, 0);
+		buffer_copy_string_len(dest, CONST_STR_LEN(""));
 		return;
 	}
 
 	force_assert('\0' == src->ptr[src->used-1]);
 
-	/* might need one character more for the '/' prefix */
-	if (src == dest) {
-		buffer_string_prepare_append(dest, 1);
-	} else {
-		buffer_string_prepare_copy(dest, buffer_string_length(src) + 1);
-	}
-
 #if defined(__WIN32) || defined(__CYGWIN__)
 	/* cygwin is treating \ and / the same, so we have to that too */
 	{
@@ -880,14 +873,15 @@ void buffer_path_simplify(buffer *dest, buffer *src)
 	while (*walk == ' ') {
 		walk++;
 	}
+	if (*walk == '.') {
+		if (walk[1] == '/' || walk[1] == '\0')
+			++walk;
+		else if (walk[1] == '.' && (walk[2] == '/' || walk[2] == '\0'))
+			walk+=2;
+	}
 
 	pre1 = 0;
 	c = *(walk++);
-	/* prefix with '/' if not already present */
-	if (c != '/') {
-		pre1 = '/';
-		*(out++) = '/';
-	}
 
 	while (c != '\0') {
 		/* assert((src != dest || out <= walk) && slash <= out); */
@@ -907,7 +901,7 @@ void buffer_path_simplify(buffer *dest, buffer *src)
 
 		if (c == '/' || c == '\0') {
 			const size_t toklen = out - slash;
-			if (toklen == 3 && pre2 == '.' && pre1 == '.') {
+			if (toklen == 3 && pre2 == '.' && pre1 == '.' && *slash == '/') {
 				/* "/../" or ("/.." at end of string) */
 				out = slash;
 				/* if there is something before "/..", there is at least one
