@@ -24,6 +24,9 @@
 #include <limits.h>
 #include <glob.h>
 
+#ifndef PATH_MAX
+#define PATH_MAX 4096
+#endif
 
 #if defined(HAVE_MYSQL) || (defined(HAVE_LDAP_H) && defined(HAVE_LBER_H) && defined(HAVE_LIBLDAP) && defined(HAVE_LIBLBER))
 static void config_warn_authn_module (server *srv, const char *module, size_t len) {
@@ -175,6 +178,7 @@ static int config_insert(server *srv) {
 	int ret = 0;
 	buffer *stat_cache_string;
 	array *http_parseopts;
+	unsigned int chunk_sz = 0;
 
 	config_values_t cv[] = {
 		{ "server.bind",                       NULL, T_CONFIG_STRING,  T_CONFIG_SCOPE_SERVER     }, /* 0 */
@@ -235,8 +239,8 @@ static int config_insert(server *srv) {
 		{ "server.network-backend",            NULL, T_CONFIG_STRING,  T_CONFIG_SCOPE_SERVER     }, /* 44 */
 		{ "server.upload-dirs",                NULL, T_CONFIG_ARRAY,   T_CONFIG_SCOPE_SERVER     }, /* 45 */
 		{ "server.core-files",                 NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER     }, /* 46 */
-		{ "unused-slot-moved-to-mod-openssl",  NULL, T_CONFIG_STRING,  T_CONFIG_SCOPE_CONNECTION }, /* 47 */
-		{ "unused-slot-moved-to-mod-openssl",  NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 48 */
+		{ "server.compat-module-load",         NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER     }, /* 47 */
+		{ "server.chunkqueue-chunk-sz",        NULL, T_CONFIG_INT,     T_CONFIG_SCOPE_SERVER     }, /* 48 */
 		{ "etag.use-inode",                    NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 49 */
 
 		{ "etag.use-mtime",                    NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 50 */
@@ -302,6 +306,8 @@ static int config_insert(server *srv) {
 	cv[44].destination = srv->srvconf.network_backend;
 	cv[45].destination = srv->srvconf.upload_tempdirs;
 	cv[46].destination = &(srv->srvconf.enable_cores);
+	cv[47].destination = &(srv->srvconf.compat_module_load);
+	cv[48].destination = &chunk_sz;
 
 	cv[52].destination = &(srv->srvconf.reject_expect_100_with_417);
 	cv[55].destination = srv->srvconf.breakagelog_file;
@@ -527,6 +533,10 @@ static int config_insert(server *srv) {
 			srv->srvconf.log_request_header_on_error = 1;
 	}
 
+	if (0 != chunk_sz) {
+		chunkqueue_set_chunk_size(chunk_sz);
+	}
+
 	if (0 != stat_cache_choose_engine(srv, stat_cache_string)) {
 		ret = HANDLER_ERROR;
 	}
@@ -542,7 +552,7 @@ static int config_insert(server *srv) {
 		log_error_write(srv, __FILE__, __LINE__, "s",
 				"unexpected value for server.modules; expected list of \"mod_xxxxxx\" strings");
 		ret = HANDLER_ERROR;
-	} else {
+	} else if (srv->srvconf.compat_module_load) {
 		data_string *ds;
 		int prepend_mod_indexfile = 1;
 		int append_mod_dirlisting = 1;
@@ -1365,39 +1375,28 @@ int config_parse_file(server *srv, config_t *context, const char *fn) {
 	return ret;
 }
 
-static char* getCWD(void) {
-	char *s, *s1;
-	size_t len;
-#ifdef PATH_MAX
-	len = PATH_MAX;
-#else
-	len = 4096;
-#endif
+#ifdef __CYGWIN__
 
-	s = malloc(len);
-	if (!s) return NULL;
-	while (NULL == getcwd(s, len)) {
-		if (errno != ERANGE || SSIZE_MAX - len < len) {
-			free(s);
-			return NULL;
-		}
-		len *= 2;
-		s1 = realloc(s, len);
-		if (!s1) {
-			free(s);
-			return NULL;
-		}
-		s = s1;
-	}
-	return s;
+static char* getCWD(char *buf, size_t sz) {
+    if (NULL == getcwd(buf, sz)) {
+        return NULL;
+    }
+    for (size_t i = 0; buf[i]; ++i) {
+        if (buf[i] == '\\') buf[i] = '/';
+    }
+    return buf;
 }
 
+#define getcwd(buf, sz) getCWD((buf),(sz))
+
+#endif /* __CYGWIN__ */
+
 int config_parse_cmd(server *srv, config_t *context, const char *cmd) {
 	int ret = 0;
-	char *oldpwd;
 	int fds[2];
+	char oldpwd[PATH_MAX];
 
-	if (NULL == (oldpwd = getCWD())) {
+	if (NULL == getcwd(oldpwd, sizeof(oldpwd))) {
 		log_error_write(srv, __FILE__, __LINE__, "s",
 			"cannot get cwd", strerror(errno));
 		return -1;
@@ -1407,7 +1406,6 @@ int config_parse_cmd(server *srv, config_t *context, const char *cmd) {
 		if (0 != chdir(context->basedir->ptr)) {
 			log_error_write(srv, __FILE__, __LINE__, "sbs",
 				"cannot change directory to", context->basedir, strerror(errno));
-			free(oldpwd);
 			return -1;
 		}
 	}
@@ -1438,7 +1436,7 @@ int config_parse_cmd(server *srv, config_t *context, const char *cmd) {
 			pid_t wpid;
 			int wstatus;
 			buffer *out = srv->tmp_buf;
-			buffer_string_set_length(out, 0);
+			buffer_clear(out);
 			close(fds[1]);
 			fds[1] = -1;
 			do {
@@ -1460,7 +1458,7 @@ int config_parse_cmd(server *srv, config_t *context, const char *cmd) {
 			}
 			if (0 != wstatus) {
 				log_error_write(srv, __FILE__, __LINE__, "SSsd",
-						"commaned \"", cmd, "\" exited non-zero:", WEXITSTATUS(wstatus));
+						"command \"", cmd, "\" exited non-zero:", WEXITSTATUS(wstatus));
 				ret = -1;
 			}
 
@@ -1481,7 +1479,6 @@ int config_parse_cmd(server *srv, config_t *context, const char *cmd) {
 			"cannot change directory to", oldpwd, strerror(errno));
 		ret = -1;
 	}
-	free(oldpwd);
 	return ret;
 }
 
@@ -1529,8 +1526,8 @@ int config_read(server *srv, const char *fn) {
 	*array_get_int_ptr(dc->value, CONST_STR_LEN("var.PID")) = getpid();
 
 	dcwd = srv->tmp_buf;
-	buffer_string_prepare_copy(dcwd, 4095);
-	if (NULL != getcwd(dcwd->ptr, 4095)) {
+	buffer_string_prepare_copy(dcwd, PATH_MAX-1);
+	if (NULL != getcwd(dcwd->ptr, buffer_string_space(dcwd)+1)) {
 		buffer_commit(dcwd, strlen(dcwd->ptr));
 		array_set_key_value(dc->value, CONST_STR_LEN("var.CWD"), CONST_BUF_LEN(dcwd));
 	}
@@ -1585,11 +1582,10 @@ int config_set_defaults(server *srv) {
 	if (srv->srvconf.upload_tempdirs->used) {
 		buffer * const b = srv->tmp_buf;
 		size_t len;
+		buffer_clear(b);
 		if (!buffer_string_is_empty(srv->srvconf.changeroot)) {
 			buffer_copy_buffer(b, srv->srvconf.changeroot);
 			buffer_append_slash(b);
-		} else {
-			buffer_reset(b);
 		}
 		len = buffer_string_length(b);
 
