@@ -30,7 +30,7 @@ int http_response_redirect_to_directory(server *srv, connection *con) {
 
 	buffer_copy_buffer(o, con->uri.scheme);
 	buffer_append_string_len(o, CONST_STR_LEN("://"));
-	if (!buffer_is_empty(con->uri.authority)) {
+	if (!buffer_string_is_empty(con->uri.authority)) {
 		buffer_append_string_buffer(o, con->uri.authority);
 	} else {
 		/* get the name of the currently connected socket */
@@ -98,7 +98,7 @@ buffer * strftime_cache_get(server *srv, time_t last_mod) {
 
 	srv->mtime_cache[i].mtime = last_mod;
 	tm = gmtime(&(srv->mtime_cache[i].mtime));
-	buffer_string_set_length(srv->mtime_cache[i].str, 0);
+	buffer_clear(srv->mtime_cache[i].str);
 	buffer_append_strftime(srv->mtime_cache[i].str, "%a, %d %b %Y %H:%M:%S GMT", tm);
 
 	return srv->mtime_cache[i].str;
@@ -189,12 +189,12 @@ int http_response_handle_cachable(server *srv, connection *con, buffer *mtime) {
 void http_response_body_clear (connection *con, int preserve_length) {
     con->response.send_chunked = 0;
     if (con->response.htags & HTTP_HEADER_TRANSFER_ENCODING) {
-        http_header_response_set(con, HTTP_HEADER_TRANSFER_ENCODING, CONST_STR_LEN("Transfer-Encoding"), CONST_STR_LEN(""));
+        http_header_response_unset(con, HTTP_HEADER_TRANSFER_ENCODING, CONST_STR_LEN("Transfer-Encoding"));
     }
     if (!preserve_length) { /* preserve for HEAD responses and no-content responses (204, 205, 304) */
         con->response.content_length = -1;
         if (con->response.htags & HTTP_HEADER_CONTENT_LENGTH) {
-            http_header_response_set(con, HTTP_HEADER_CONTENT_LENGTH, CONST_STR_LEN("Content-Length"), CONST_STR_LEN(""));
+            http_header_response_unset(con, HTTP_HEADER_CONTENT_LENGTH, CONST_STR_LEN("Content-Length"));
         }
     }
     chunkqueue_reset(con->write_queue);
@@ -325,8 +325,7 @@ static int http_response_parse_range(server *srv, connection *con, buffer *path,
 		if (!error) {
 			if (multipart) {
 				/* write boundary-header */
-				buffer *b = buffer_init();
-
+				buffer *b = srv->tmp_buf;
 				buffer_copy_string_len(b, CONST_STR_LEN("\r\n--"));
 				buffer_append_string_len(b, boundary, sizeof(boundary)-1);
 
@@ -347,8 +346,7 @@ static int http_response_parse_range(server *srv, connection *con, buffer *path,
 				buffer_append_string_len(b, CONST_STR_LEN("\r\n\r\n"));
 
 				con->response.content_length += buffer_string_length(b);
-				chunkqueue_append_buffer(con->write_queue, b);
-				buffer_free(b);
+				chunkqueue_append_mem(con->write_queue, CONST_BUF_LEN(b));
 			}
 
 			chunkqueue_append_file(con->write_queue, path, start, end - start + 1);
@@ -361,15 +359,13 @@ static int http_response_parse_range(server *srv, connection *con, buffer *path,
 
 	if (multipart) {
 		/* add boundary end */
-		buffer *b = buffer_init();
-
+		buffer *b = srv->tmp_buf;
 		buffer_copy_string_len(b, "\r\n--", 4);
 		buffer_append_string_len(b, boundary, sizeof(boundary)-1);
 		buffer_append_string_len(b, "--\r\n", 4);
 
 		con->response.content_length += buffer_string_length(b);
-		chunkqueue_append_buffer(con->write_queue, b);
-		buffer_free(b);
+		chunkqueue_append_mem(con->write_queue, CONST_BUF_LEN(b));
 
 		/* set header-fields */
 
@@ -553,11 +549,20 @@ static void http_response_xsendfile (server *srv, connection *con, buffer *path,
 	 * determined by open(), fstat() to reduces race conditions if the file
 	 * is modified between stat() (stat_cache_get_entry()) and open(). */
 	if (con->response.htags & HTTP_HEADER_CONTENT_LENGTH) {
-		http_header_response_set(con, HTTP_HEADER_CONTENT_LENGTH, CONST_STR_LEN("Content-Length"), CONST_STR_LEN(""));
+		http_header_response_unset(con, HTTP_HEADER_CONTENT_LENGTH, CONST_STR_LEN("Content-Length"));
 		con->response.content_length = -1;
 	}
 
 	buffer_urldecode_path(path);
+	if (!buffer_is_valid_UTF8(path)) {
+		log_error_write(srv, __FILE__, __LINE__, "sb",
+				"X-Sendfile invalid UTF-8 after url-decode:", path);
+		if (con->http_status < 400) {
+			con->http_status = 502;
+			con->mode = DIRECT;
+		}
+		return;
+	}
 	buffer_path_simplify(path, path);
 	if (con->conf.force_lowercase_filenames) {
 		buffer_to_lower(path);
@@ -605,7 +610,7 @@ static void http_response_xsendfile2(server *srv, connection *con, const buffer
 
     /* reset Content-Length, if set by backend */
     if (con->response.htags & HTTP_HEADER_CONTENT_LENGTH) {
-        http_header_response_set(con, HTTP_HEADER_CONTENT_LENGTH, CONST_STR_LEN("Content-Length"), CONST_STR_LEN(""));
+        http_header_response_unset(con, HTTP_HEADER_CONTENT_LENGTH, CONST_STR_LEN("Content-Length"));
         con->response.content_length = -1;
     }
 
@@ -631,6 +636,12 @@ static void http_response_xsendfile2(server *srv, connection *con, const buffer
         for (pos = ++range; *pos && *pos != ' ' && *pos != ','; pos++) ;
 
         buffer_urldecode_path(b);
+        if (!buffer_is_valid_UTF8(b)) {
+            log_error_write(srv, __FILE__, __LINE__, "sb",
+                            "X-Sendfile2 invalid UTF-8 after url-decode:", b);
+            con->http_status = 502;
+            break;
+        }
         buffer_path_simplify(b, b);
         if (con->conf.force_lowercase_filenames) {
             buffer_to_lower(b);
@@ -929,6 +940,7 @@ static int http_response_process_headers(server *srv, connection *con, http_resp
                 status = strtol(value, NULL, 10);
                 if (status >= 100 && status < 1000) {
                     con->http_status = status;
+                    status_is_set = 1;
                 } else {
                     con->http_status = 502;
                     con->mode = DIRECT;
@@ -1105,14 +1117,16 @@ handler_t http_response_parse_headers(server *srv, connection *con, http_respons
         if (opts->backend == BACKEND_FASTCGI
             && NULL != (vb = http_header_response_get(con, HTTP_HEADER_OTHER, CONST_STR_LEN("X-Sendfile2")))) {
             http_response_xsendfile2(srv, con, vb, opts->xsendfile_docroot);
-            buffer_reset(vb); /*(do not send to client)*/
+            /* http_header_response_unset() shortcut for HTTP_HEADER_OTHER */
+            buffer_clear(vb); /*(do not send to client)*/
             if (con->mode == DIRECT) con->file_started = 0;
             return HANDLER_FINISHED;
         } else if (NULL != (vb = http_header_response_get(con, HTTP_HEADER_OTHER, CONST_STR_LEN("X-Sendfile")))
                    || (opts->backend == BACKEND_FASTCGI /* X-LIGHTTPD-send-file is deprecated; historical for fastcgi */
                        && NULL != (vb = http_header_response_get(con, HTTP_HEADER_OTHER, CONST_STR_LEN("X-LIGHTTPD-send-file"))))) {
             http_response_xsendfile(srv, con, vb, opts->xsendfile_docroot);
-            buffer_reset(vb); /*(do not send to client)*/
+            /* http_header_response_unset() shortcut for HTTP_HEADER_OTHER */
+            buffer_clear(vb); /*(do not send to client)*/
             if (con->mode == DIRECT) con->file_started = 0;
             return HANDLER_FINISHED;
         }
@@ -1133,12 +1147,13 @@ handler_t http_response_read(server *srv, connection *con, http_response_opts *o
     while (1) {
         ssize_t n;
         size_t avail = buffer_string_space(b);
-        unsigned int toread = 4096;
+        unsigned int toread = 0;
 
         if (0 == fdevent_ioctl_fionread(fd, opts->fdfmt, (int *)&toread)) {
             if (avail < toread) {
-                if (toread < 4096)
-                    toread = 4096;
+                size_t blen = buffer_string_length(b);
+                if (toread + blen < 4096)
+                    toread = 4095 - blen;
                 else if (toread > MAX_READ_LIMIT)
                     toread = MAX_READ_LIMIT;
             }
@@ -1153,10 +1168,14 @@ handler_t http_response_read(server *srv, connection *con, http_response_opts *o
                           & FDEVENT_STREAM_RESPONSE_POLLRDHUP))
                         return HANDLER_GO_ON;/*optimistic read; data not ready*/
                 }
-                toread = 4096; /* let read() below indicate if EOF or EAGAIN */
+                if (0 == avail) /* let read() below indicate if EOF or EAGAIN */
+                    toread = 1024;
               #endif
             }
         }
+        else if (avail < 1024) {
+            toread = 4095 - avail;
+        }
 
         if (con->conf.stream_response_body & FDEVENT_STREAM_RESPONSE_BUFMIN) {
             off_t cqlen = chunkqueue_length(con->write_queue);
@@ -1220,14 +1239,14 @@ handler_t http_response_read(server *srv, connection *con, http_response_opts *o
             handler_t rc = http_response_parse_headers(srv, con, opts, b);
             if (rc != HANDLER_GO_ON) return rc;
             /* accumulate response in b until headers completed (or error) */
-            if (con->file_started) buffer_string_set_length(b, 0);
+            if (con->file_started) buffer_clear(b);
         } else {
             if (0 != http_chunk_append_buffer(srv, con, b)) {
                 /* error writing to tempfile;
                  * truncate response or send 500 if nothing sent yet */
                 return HANDLER_ERROR;
             }
-            buffer_string_set_length(b, 0);
+            buffer_clear(b);
         }
 
         if ((con->conf.stream_response_body & FDEVENT_STREAM_RESPONSE_BUFMIN)
