@@ -3,6 +3,7 @@
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
+#include "http_header.h"
 
 #include "plugin.h"
 
@@ -201,13 +202,7 @@ URIHANDLER_FUNC(mod_indexfile_subrequest) {
 		if (ds->value && ds->value->ptr[0] == '/') {
 			/* replace uri.path */
 			buffer_copy_buffer(con->uri.path, ds->value);
-
-			if (NULL == (ds = (data_string *)array_get_unused_element(con->environment, TYPE_STRING))) {
-				ds = data_string_init();
-			}
-			buffer_copy_string_len(ds->key, CONST_STR_LEN("PATH_TRANSLATED_DIRINDEX"));
-			buffer_copy_buffer(ds->value, con->physical.path);
-			array_insert_unique(con->environment, (data_unset *)ds);
+			http_header_env_set(con, CONST_STR_LEN("PATH_TRANSLATED_DIRINDEX"), CONST_BUF_LEN(con->physical.path));
 		} else {
 			/* append to uri.path the relative path to index file (/ -> /index.php) */
 			buffer_append_string_buffer(con->uri.path, ds->value);
