@@ -327,6 +327,7 @@ network_ssl_servername_callback (SSL *ssl, int *al, server *srv)
     const char *servername;
     handler_ctx *hctx = (handler_ctx *) SSL_get_app_data(ssl);
     connection *con = hctx->con;
+    size_t len;
     UNUSED(al);
 
     buffer_copy_string(con->uri.scheme, "https");
@@ -340,8 +341,14 @@ network_ssl_servername_callback (SSL *ssl, int *al, server *srv)
 #endif
         return SSL_TLSEXT_ERR_NOACK;
     }
+    len = strlen(servername);
+    if (len >= 1024) { /*(expecting < 256)*/
+        log_error_write(srv, __FILE__, __LINE__, "sss", "SSL:",
+                        "SNI name too long", servername);
+        return SSL_TLSEXT_ERR_ALERT_FATAL;
+    }
     /* use SNI to patch mod_openssl config and then reset COMP_HTTP_HOST */
-    buffer_copy_string(con->uri.authority, servername);
+    buffer_copy_string_len(con->uri.authority, servername, len);
     buffer_to_lower(con->uri.authority);
 
     con->conditional_is_valid[COMP_HTTP_SCHEME] = 1;
@@ -350,7 +357,7 @@ network_ssl_servername_callback (SSL *ssl, int *al, server *srv)
     /* reset COMP_HTTP_HOST so that conditions re-run after request hdrs read */
     /*(done in response.c:config_cond_cache_reset() after request hdrs read)*/
     /*config_cond_cache_reset_item(con, COMP_HTTP_HOST);*/
-    /*buffer_reset(con->uri.authority);*/
+    /*buffer_clear(con->uri.authority);*/
 
     if (NULL == hctx->conf.ssl_pemfile_x509
         || NULL == hctx->conf.ssl_pemfile_pkey) {
@@ -711,7 +718,14 @@ network_init_ssl (server *srv, void *p_d)
 
         if (buffer_string_is_empty(s->ssl_pemfile) || !s->ssl_enabled) continue;
 
-        if (NULL == (s->ssl_ctx = SSL_CTX_new(SSLv23_server_method()))) {
+      #if OPENSSL_VERSION_NUMBER >= 0x10100000L
+        s->ssl_ctx = (!s->ssl_use_sslv2 && !s->ssl_use_sslv3)
+          ? SSL_CTX_new(TLS_server_method())
+          : SSL_CTX_new(SSLv23_server_method());
+      #else
+        s->ssl_ctx = SSL_CTX_new(SSLv23_server_method());
+      #endif
+        if (NULL == s->ssl_ctx) {
             log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:",
                             ERR_error_string(ERR_get_error(), NULL));
             return -1;
@@ -1423,8 +1437,9 @@ connection_read_cq_ssl (server *srv, connection *con,
 
     ERR_clear_error();
     do {
-        chunkqueue_get_memory(con->read_queue, &mem, &mem_len, 0,
-                              SSL_pending(hctx->ssl));
+        len = SSL_pending(hctx->ssl);
+        mem_len = len < 2048 ? 2048 : (size_t)len;
+        mem = chunkqueue_get_memory(con->read_queue, &mem_len);
 #if 0
         /* overwrite everything with 0 */
         memset(mem, 0, mem_len);
@@ -1580,6 +1595,10 @@ CONNECTION_FUNC(mod_openssl_handle_con_accept)
 }
 
 
+static void
+mod_openssl_close_notify(server *srv, handler_ctx *hctx);
+
+
 CONNECTION_FUNC(mod_openssl_handle_con_shut_wr)
 {
     plugin_data *p = p_d;
@@ -1587,6 +1606,16 @@ CONNECTION_FUNC(mod_openssl_handle_con_shut_wr)
     if (NULL == hctx) return HANDLER_GO_ON;
 
     if (SSL_is_init_finished(hctx->ssl)) {
+        mod_openssl_close_notify(srv, hctx);
+    }
+
+    return HANDLER_GO_ON;
+}
+
+
+static void
+mod_openssl_close_notify(server *srv, handler_ctx *hctx)
+{
         int ret, ssl_r;
         unsigned long err;
         ERR_clear_error();
@@ -1600,6 +1629,33 @@ CONNECTION_FUNC(mod_openssl_handle_con_shut_wr)
              * FIXME: wait for fdevent and call SSL_shutdown again
              *
              */
+
+            /* Drain SSL read buffers in case pending records need processing.
+             * Limit to reading 16k to avoid denial of service when the CPU
+             * processing TLS is slower than arrival speed of TLS data packets.
+             *
+             * references:
+             *
+             * "New session ticket breaks bidirectional shutdown of TLS 1.3 connection"
+             * https://github.com/openssl/openssl/issues/6262
+             *
+             * The peer is still allowed to send data after receiving the
+             * "close notify" event. If the peer did send data it need to be
+             * processed by calling SSL_read() before calling SSL_shutdown() a
+             * second time. SSL_read() will indicate the end of the peer data by
+             * returning <= 0 and SSL_get_error() returning
+             * SSL_ERROR_ZERO_RETURN. It is recommended to call SSL_read()
+             * between SSL_shutdown() calls.
+             *
+             * Additional discussion in "Auto retry in shutdown"
+             * https://github.com/openssl/openssl/pull/6340
+             */
+            err = 0;
+            do {
+                char buf[4096];
+                ret = SSL_read(hctx->ssl, buf, (int)sizeof(buf));
+            } while (ret > 0 && (err += (unsigned long)ret) < 16384);
+
             ERR_clear_error();
             if (-1 != (ret = SSL_shutdown(hctx->ssl))) break;
 
@@ -1647,9 +1703,6 @@ CONNECTION_FUNC(mod_openssl_handle_con_shut_wr)
             }
         }
         ERR_clear_error();
-    }
-
-    return HANDLER_GO_ON;
 }
 
 
