@@ -535,8 +535,11 @@ static void proxy_set_Forwarded(connection *con, const unsigned int flags) {
         http_header_request_set(con, HTTP_HEADER_FORWARDED,
                                 CONST_STR_LEN("Forwarded"),
                                 CONST_STR_LEN("x")); /*(must not be blank for _get below)*/
+      #ifdef __COVERITY__
+        force_assert(NULL != b); /*(not NULL because created directly above)*/
+      #endif
         b = http_header_request_get(con, HTTP_HEADER_FORWARDED, CONST_STR_LEN("Forwarded"));
-        buffer_string_set_length(b, 0);
+        buffer_clear(b);
         if (NULL != xff) {
             /* use X-Forwarded-For contents to seed Forwarded */
             char *s = xff->ptr;
@@ -708,12 +711,11 @@ static void proxy_set_Forwarded(connection *con, const unsigned int flags) {
 static handler_t proxy_create_env(server *srv, gw_handler_ctx *gwhctx) {
 	handler_ctx *hctx = (handler_ctx *)gwhctx;
 	connection *con = hctx->gw.remote_conn;
-	buffer *b = buffer_init();
 	const int remap_headers = (NULL != hctx->conf.header.urlpaths
 				   || NULL != hctx->conf.header.hosts_request);
 	const int upgrade = hctx->conf.header.upgrade
 	    && (NULL != http_header_request_get(con, HTTP_HEADER_UPGRADE, CONST_STR_LEN("Upgrade")));
-	buffer_string_prepare_copy(b, 8192-1);
+	buffer * const b = chunkqueue_prepend_buffer_open_sz(hctx->gw.wb, (size_t)(con->read_queue->bytes_out - hctx->gw.wb->bytes_in));
 
 	/* build header */
 
@@ -793,6 +795,12 @@ static handler_t proxy_create_env(server *srv, gw_handler_ctx *gwhctx) {
 		vlen = buffer_string_length(ds->value);
 		if (0 == vlen) continue;
 
+		if (buffer_string_space(b) < klen + vlen + 4) {
+			size_t extend = b->size * 2 - buffer_string_length(b);
+			extend = extend > klen + vlen + 4 ? extend : klen + vlen + 4 + 4095;
+			buffer_string_prepare_append(b, extend);
+		}
+
 		buffer_append_string_len(b, ds->key->ptr, klen);
 		buffer_append_string_len(b, CONST_STR_LEN(": "));
 		buffer_append_string_len(b, ds->value->ptr, vlen);
@@ -832,8 +840,7 @@ static handler_t proxy_create_env(server *srv, gw_handler_ctx *gwhctx) {
 		buffer_append_string_len(b, CONST_STR_LEN("Connection: close, upgrade\r\n\r\n"));
 
 	hctx->gw.wb_reqlen = buffer_string_length(b);
-	chunkqueue_append_buffer(hctx->gw.wb, b);
-	buffer_free(b);
+	chunkqueue_prepend_buffer_commit(hctx->gw.wb);
 
 	if (con->request.content_length) {
 		chunkqueue_append_chunkqueue(hctx->gw.wb, con->request_content_queue);
@@ -932,9 +939,8 @@ static handler_t proxy_response_headers(server *srv, connection *con, struct htt
             /* preserve prior questionable behavior; likely broken behavior
              * anyway if backend thinks connection is being upgraded but client
              * does not receive Connection: upgrade */
-            http_header_response_set(con, HTTP_HEADER_UPGRADE,
-                                     CONST_STR_LEN("Upgrade"),
-                                     CONST_STR_LEN(""));
+            http_header_response_unset(con, HTTP_HEADER_UPGRADE,
+                                       CONST_STR_LEN("Upgrade"))
           #endif
         }
     }
@@ -976,7 +982,7 @@ static handler_t mod_proxy_check_extension(server *srv, connection *con, void *p
 	if (con->mode == p->id) {
 		handler_ctx *hctx = con->plugin_ctx[p->id];
 		hctx->gw.create_env = proxy_create_env;
-		hctx->gw.response = buffer_init();
+		hctx->gw.response = chunk_buffer_acquire();
 		hctx->gw.opts.backend = BACKEND_PROXY;
 		hctx->gw.opts.pdata = hctx;
 		hctx->gw.opts.headers = proxy_response_headers;
