@@ -171,7 +171,7 @@ static int build_ssi_cgi_vars(server *srv, connection *con, handler_ctx *p) {
 		memset(vb_auth, 0, sizeof(buffer));
 	}
 
-	array_reset(p->ssi_cgi_env);
+	array_reset_data_strings(p->ssi_cgi_env);
 
 	if (0 != http_cgi_headers(srv, con, &opts, ssi_env_add, p->ssi_cgi_env)) {
 		con->http_status = 400;
@@ -371,7 +371,7 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 		case SSI_ECHO_USER_NAME: {
 			struct passwd *pw;
 
-			b = buffer_init();
+			b = srv->tmp_buf;
 #ifdef HAVE_PWD_H
 			if (NULL == (pw = getpwuid(st->st_uid))) {
 				buffer_copy_int(b, st->st_uid);
@@ -381,8 +381,7 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 #else
 			buffer_copy_int(b, st->st_uid);
 #endif
-			chunkqueue_append_buffer(con->write_queue, b);
-			buffer_free(b);
+			chunkqueue_append_mem(con->write_queue, CONST_BUF_LEN(b));
 			break;
 		}
 		case SSI_ECHO_LAST_MODIFIED: {
@@ -512,6 +511,11 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 
 			buffer_copy_string(srv->tmp_buf, file_path);
 			buffer_urldecode_path(srv->tmp_buf);
+			if (!buffer_is_valid_UTF8(srv->tmp_buf)) {
+				log_error_write(srv, __FILE__, __LINE__, "sb",
+						"SSI invalid UTF-8 after url-decode:", srv->tmp_buf);
+				break;
+			}
 			buffer_path_simplify(srv->tmp_buf, srv->tmp_buf);
 			buffer_append_string_buffer(p->stat_fn, srv->tmp_buf);
 		} else {
@@ -519,17 +523,22 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 			size_t remain;
 
 			if (virt_path[0] == '/') {
-				buffer_copy_string(p->stat_fn, virt_path);
+				buffer_copy_string(srv->tmp_buf, virt_path);
 			} else {
 				/* there is always a / */
 				sl = strrchr(con->uri.path->ptr, '/');
 
-				buffer_copy_string_len(p->stat_fn, con->uri.path->ptr, sl - con->uri.path->ptr + 1);
-				buffer_append_string(p->stat_fn, virt_path);
+				buffer_copy_string_len(srv->tmp_buf, con->uri.path->ptr, sl - con->uri.path->ptr + 1);
+				buffer_append_string(srv->tmp_buf, virt_path);
 			}
 
-			buffer_urldecode_path(p->stat_fn);
-			buffer_path_simplify(srv->tmp_buf, p->stat_fn);
+			buffer_urldecode_path(srv->tmp_buf);
+			if (!buffer_is_valid_UTF8(srv->tmp_buf)) {
+				log_error_write(srv, __FILE__, __LINE__, "sb",
+						"SSI invalid UTF-8 after url-decode:", srv->tmp_buf);
+				break;
+			}
+			buffer_path_simplify(srv->tmp_buf, srv->tmp_buf);
 
 			/* we have an uri */
 
@@ -580,7 +589,7 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 
 			switch (ssicmd) {
 			case SSI_FSIZE:
-				b = buffer_init();
+				b = srv->tmp_buf;
 				if (p->sizefmt) {
 					int j = 0;
 					const char *abr[] = { " B", " kB", " MB", " GB", " TB", NULL };
@@ -594,8 +603,7 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 				} else {
 					buffer_copy_int(b, stb.st_size);
 				}
-				chunkqueue_append_buffer(con->write_queue, b);
-				buffer_free(b);
+				chunkqueue_append_mem(con->write_queue, CONST_BUF_LEN(b));
 				break;
 			case SSI_FLASTMOD:
 				if (0 == strftime(buf, sizeof(buf), p->timefmt->ptr, localtime(&t))) {
@@ -719,7 +727,8 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 	case SSI_PRINTENV:
 		if (p->if_is_false) break;
 
-		b = buffer_init();
+		b = srv->tmp_buf;
+		buffer_clear(b);
 		for (i = 0; i < p->ssi_vars->used; i++) {
 			data_string *ds = (data_string *)p->ssi_vars->data[p->ssi_vars->sorted[i]];
 
@@ -736,9 +745,7 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 			buffer_append_string_encoded(b, CONST_BUF_LEN(ds->value), ENCODING_MINIMAL_XML);
 			buffer_append_string_len(b, CONST_STR_LEN("\n"));
 		}
-		chunkqueue_append_buffer(con->write_queue, b);
-		buffer_free(b);
-
+		chunkqueue_append_mem(con->write_queue, CONST_BUF_LEN(b));
 		break;
 	case SSI_EXEC: {
 		const char *cmd = NULL;
@@ -1195,8 +1202,8 @@ static int mod_ssi_handle_request(server *srv, connection *con, handler_ctx *p)
 
 	/* get a stream to the file */
 
-	array_reset(p->ssi_vars);
-	array_reset(p->ssi_cgi_env);
+	array_reset_data_strings(p->ssi_vars);
+	array_reset_data_strings(p->ssi_cgi_env);
 	buffer_copy_string_len(p->timefmt, CONST_STR_LEN("%a, %d %b %Y %H:%M:%S %Z"));
 	build_ssi_cgi_vars(srv, con, p);
 
