@@ -102,7 +102,6 @@ INIT_FUNC(mod_webdav_init) {
 	p->tmp_buf = buffer_init();
 
 	p->uri.scheme = buffer_init();
-	p->uri.path_raw = buffer_init();
 	p->uri.path = buffer_init();
 	p->uri.authority = buffer_init();
 
@@ -154,7 +153,6 @@ FREE_FUNC(mod_webdav_free) {
 	}
 
 	buffer_free(p->uri.scheme);
-	buffer_free(p->uri.path_raw);
 	buffer_free(p->uri.path);
 	buffer_free(p->uri.authority);
 
@@ -592,6 +590,7 @@ static int webdav_delete_dir(server *srv, connection *con, handler_ctx *hctx, ph
 
 		while(NULL != (de = readdir(dir))) {
 			struct stat st;
+			size_t nlen;
 
 			if ((de->d_name[0] == '.' && de->d_name[1] == '\0')  ||
 			  (de->d_name[0] == '.' && de->d_name[1] == '.' && de->d_name[2] == '\0')) {
@@ -599,13 +598,12 @@ static int webdav_delete_dir(server *srv, connection *con, handler_ctx *hctx, ph
 				/* ignore the parent dir */
 			}
 
+			nlen = strlen(de->d_name);
 			buffer_copy_buffer(d.path, dst->path);
-			buffer_append_slash(d.path);
-			buffer_append_string(d.path, de->d_name);
+			buffer_append_path_len(d.path, de->d_name, nlen);
 
 			buffer_copy_buffer(d.rel_path, dst->rel_path);
-			buffer_append_slash(d.rel_path);
-			buffer_append_string(d.rel_path, de->d_name);
+			buffer_append_path_len(d.rel_path, de->d_name, nlen);
 
 			/* stat and unlink afterwards */
 			if (-1 == stat(d.path->ptr, &st)) {
@@ -772,27 +770,25 @@ static int webdav_copy_dir(server *srv, connection *con, handler_ctx *hctx, phys
 
 		while (NULL != (de = readdir(srcdir))) {
 			struct stat st;
+			size_t nlen;
 
 			if ((de->d_name[0] == '.' && de->d_name[1] == '\0')
 				|| (de->d_name[0] == '.' && de->d_name[1] == '.' && de->d_name[2] == '\0')) {
 				continue;
 			}
 
+			nlen = strlen(de->d_name);
 			buffer_copy_buffer(s.path, src->path);
-			buffer_append_slash(s.path);
-			buffer_append_string(s.path, de->d_name);
+			buffer_append_path_len(s.path, de->d_name, nlen);
 
 			buffer_copy_buffer(d.path, dst->path);
-			buffer_append_slash(d.path);
-			buffer_append_string(d.path, de->d_name);
+			buffer_append_path_len(d.path, de->d_name, nlen);
 
 			buffer_copy_buffer(s.rel_path, src->rel_path);
-			buffer_append_slash(s.rel_path);
-			buffer_append_string(s.rel_path, de->d_name);
+			buffer_append_path_len(s.rel_path, de->d_name, nlen);
 
 			buffer_copy_buffer(d.rel_path, dst->rel_path);
-			buffer_append_slash(d.rel_path);
-			buffer_append_string(d.rel_path, de->d_name);
+			buffer_append_path_len(d.rel_path, de->d_name, nlen);
 
 			if (-1 == stat(s.path->ptr, &st)) {
 				/* why ? */
@@ -983,7 +979,7 @@ static int webdav_get_live_property(server *srv, connection *con, handler_ctx *h
 			buffer_append_string_len(b, CONST_STR_LEN("<D:getetag>"));
 			buffer_append_string_buffer(b, con->physical.etag);
 			buffer_append_string_len(b, CONST_STR_LEN("</D:getetag>"));
-			buffer_reset(con->physical.etag);
+			buffer_clear(con->physical.etag);
 			found = 1;
 	      #ifdef USE_LOCKS
 		} else if (0 == strcmp(prop_name, "lockdiscovery")) {
@@ -1132,7 +1128,7 @@ static int webdav_parse_chunkqueue(server *srv, connection *con, handler_ctx *hc
 				data = c->file.mmap.start + c->offset;
 			} else {
 				if (-1 == c->file.fd &&  /* open the file if not already open */
-				    -1 == (c->file.fd = fdevent_open_cloexec(c->file.name->ptr, O_RDONLY, 0))) {
+				    -1 == (c->file.fd = fdevent_open_cloexec(c->mem->ptr, O_RDONLY, 0))) {
 					log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
 
 					return -1;
@@ -1151,7 +1147,7 @@ static int webdav_parse_chunkqueue(server *srv, connection *con, handler_ctx *hc
 					if (-1 == lseek(c->file.fd, c->file.start + c->offset, SEEK_SET)
 					    || 0 > (rd = read(c->file.fd, data, weHave))) {
 						log_error_write(srv, __FILE__, __LINE__, "ssbd", "lseek/read failed: ",
-								strerror(errno), c->file.name, c->file.fd);
+								strerror(errno), c->mem, c->file.fd);
 						free(data);
 						return -1;
 					}
@@ -1213,7 +1209,7 @@ static int webdav_parse_chunkqueue(server *srv, connection *con, handler_ctx *hc
 #ifdef USE_LOCKS
 static int webdav_lockdiscovery(connection *con, buffer *locktoken, const char *lockscope, const char *locktype, int depth) {
 
-	buffer *b = buffer_init();
+	buffer *b = chunkqueue_append_buffer_open(con->write_queue);
 
 	http_header_response_set(con, HTTP_HEADER_OTHER, CONST_STR_LEN("Lock-Token"), CONST_BUF_LEN(locktoken));
 
@@ -1229,8 +1225,7 @@ static int webdav_lockdiscovery(connection *con, buffer *locktoken, const char *
 	buffer_append_string_len(b,CONST_STR_LEN("</D:lockdiscovery>\n"));
 	buffer_append_string_len(b,CONST_STR_LEN("</D:prop>\n"));
 
-	chunkqueue_append_buffer(con->write_queue, b);
-	buffer_free(b);
+	chunkqueue_append_buffer_commit(con->write_queue);
 
 	return 0;
 }
@@ -1295,37 +1290,26 @@ static int webdav_has_lock(server *srv, connection *con, handler_ctx *hctx, buff
 	return has_lock;
 }
 
+static int mod_webdav_depth(connection *con) {
+    buffer *b =
+      http_header_request_get(con, HTTP_HEADER_OTHER, CONST_STR_LEN("Depth"));
+    if (NULL != b && 1 == buffer_string_length(b)) {
+        if (b->ptr[0] == '0') return 0;
+        if (b->ptr[0] == '1') return 1;
+    }
+    return -1; /* (Depth: infinity) */
+}
 
-SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
-	plugin_data *p = p_d;
-	handler_ctx *hctx = con->plugin_ctx[p->id];
+static handler_t mod_webdav_propfind(server *srv, connection *con, plugin_data *p, handler_ctx *hctx) {
 	buffer *b;
 	DIR *dir;
-	int depth = -1; /* (Depth: infinity) */
+	int depth = mod_webdav_depth(con);
 	struct stat st;
 	buffer *prop_200;
 	buffer *prop_404;
 	webdav_properties *req_props;
 	stat_cache_entry *sce = NULL;
 
-	UNUSED(srv);
-
-	if (NULL == hctx) return HANDLER_GO_ON;
-	if (!hctx->conf.enabled) return HANDLER_GO_ON;
-	/* physical path is setup */
-	if (buffer_is_empty(con->physical.path)) return HANDLER_GO_ON;
-
-	/* PROPFIND need them */
-	if (NULL != (b = http_header_request_get(con, HTTP_HEADER_OTHER, CONST_STR_LEN("Depth"))) && 1 == buffer_string_length(b)) {
-		if ('0' == *b->ptr) {
-			depth = 0;
-		} else if ('1' == *b->ptr) {
-			depth = 1;
-		}
-	} /* else treat as Depth: infinity */
-
-	switch (con->request.http_method) {
-	case HTTP_METHOD_PROPFIND:
 		/* they want to know the properties of the directory */
 		req_props = NULL;
 
@@ -1452,7 +1436,7 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 
 		http_header_response_set(con, HTTP_HEADER_CONTENT_TYPE, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/xml; charset=\"utf-8\""));
 
-		b = buffer_init();
+		b = chunkqueue_append_buffer_open(con->write_queue);
 
 		buffer_copy_string_len(b, CONST_STR_LEN("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"));
 
@@ -1514,22 +1498,21 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 				d.rel_path = buffer_init();
 
 				while(NULL != (de = readdir(dir))) {
+					size_t nlen;
 					if (de->d_name[0] == '.' && (de->d_name[1] == '\0' || (de->d_name[1] == '.' && de->d_name[2] == '\0'))) {
 						continue;
 						/* ignore the parent and target dir */
 					}
 
+					nlen = strlen(de->d_name);
 					buffer_copy_buffer(d.path, dst->path);
-					buffer_append_slash(d.path);
+					buffer_append_path_len(d.path, de->d_name, nlen);
 
 					buffer_copy_buffer(d.rel_path, dst->rel_path);
-					buffer_append_slash(d.rel_path);
+					buffer_append_path_len(d.rel_path, de->d_name, nlen);
 
-					buffer_append_string(d.path, de->d_name);
-					buffer_append_string(d.rel_path, de->d_name);
-
-					buffer_reset(prop_200);
-					buffer_reset(prop_404);
+					buffer_clear(prop_200);
+					buffer_clear(prop_404);
 
 					webdav_get_props(srv, con, hctx, &d, req_props, prop_200, prop_404);
 
@@ -1599,13 +1582,14 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 			log_error_write(srv, __FILE__, __LINE__, "sb", "XML-response-body:", b);
 		}
 
-		chunkqueue_append_buffer(con->write_queue, b);
-		buffer_free(b);
+		chunkqueue_append_buffer_commit(con->write_queue);
 
 		con->file_finished = 1;
 
 		return HANDLER_FINISHED;
-	case HTTP_METHOD_MKCOL:
+}
+
+static handler_t mod_webdav_mkcol(connection *con, plugin_data *p) {
 		if (p->conf.is_readonly) {
 			con->http_status = 403;
 			return HANDLER_FINISHED;
@@ -1640,7 +1624,11 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 		}
 
 		return HANDLER_FINISHED;
-	case HTTP_METHOD_DELETE:
+}
+
+static handler_t mod_webdav_delete(server *srv, connection *con, plugin_data *p, handler_ctx *hctx) {
+		struct stat st;
+
 		if (p->conf.is_readonly) {
 			con->http_status = 403;
 			return HANDLER_FINISHED;
@@ -1675,9 +1663,10 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 
 			if (webdav_delete_dir(srv, con, hctx, &(con->physical), multi_status_resp)) {
 				/* we got an error somewhere in between, build a 207 */
+				buffer *b;
 				http_header_response_set(con, HTTP_HEADER_CONTENT_TYPE, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/xml; charset=\"utf-8\""));
 
-				b = buffer_init();
+				b = chunkqueue_append_buffer_open(con->write_queue);
 
 				buffer_copy_string_len(b, CONST_STR_LEN("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"));
 
@@ -1691,8 +1680,7 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 					log_error_write(srv, __FILE__, __LINE__, "sb", "XML-response-body:", b);
 				}
 
-				chunkqueue_append_buffer(con->write_queue, b);
-				buffer_free(b);
+				chunkqueue_append_buffer_commit(con->write_queue);
 
 				con->http_status = 207;
 				con->file_finished = 1;
@@ -1733,7 +1721,10 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 			con->http_status = 204;
 		}
 		return HANDLER_FINISHED;
-	case HTTP_METHOD_PUT: {
+}
+
+static handler_t mod_webdav_put(server *srv, connection *con, plugin_data *p, handler_ctx *hctx) {
+		buffer *b;
 		int fd;
 		chunkqueue *cq = con->request_content_queue;
 		chunk *c;
@@ -1850,7 +1841,7 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 					data = c->file.mmap.start + c->offset;
 				} else {
 					if (-1 == c->file.fd &&  /* open the file if not already open */
-					    -1 == (c->file.fd = fdevent_open_cloexec(c->file.name->ptr, O_RDONLY, 0))) {
+					    -1 == (c->file.fd = fdevent_open_cloexec(c->mem->ptr, O_RDONLY, 0))) {
 						log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
 						close(fd);
 						return HANDLER_ERROR;
@@ -1869,7 +1860,7 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 						if (-1 == lseek(c->file.fd, c->file.start + c->offset, SEEK_SET)
 						    || 0 > (rd = read(c->file.fd, data, dlen))) {
 							log_error_write(srv, __FILE__, __LINE__, "ssbd", "lseek/read failed: ",
-									strerror(errno), c->file.name, c->file.fd);
+									strerror(errno), c->mem, c->file.fd);
 							free(data);
 							close(fd);
 							return HANDLER_ERROR;
@@ -1921,9 +1912,11 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 		}
 
 		return HANDLER_FINISHED;
-	}
-	case HTTP_METHOD_MOVE:
-	case HTTP_METHOD_COPY: {
+}
+
+static handler_t mod_webdav_copymove(server *srv, connection *con, plugin_data *p, handler_ctx *hctx) {
+		buffer *b;
+		struct stat st;
 		buffer *destination = NULL;
 		char *sep, *sep2, *start;
 		int overwrite = 1;
@@ -1964,10 +1957,6 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 		 * - the query string is thrown away
 		 *  */
 
-		buffer_reset(p->uri.scheme);
-		buffer_reset(p->uri.path_raw);
-		buffer_reset(p->uri.authority);
-
 		start = destination->ptr;
 
 		if (NULL == (sep = strstr(start, "://"))) {
@@ -1992,9 +1981,9 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 
 		if (NULL == (sep = strchr(start, '?'))) {
 			/* no query string, good */
-			buffer_copy_string(p->uri.path_raw, start);
+			buffer_copy_string(p->uri.path, start);
 		} else {
-			buffer_copy_string_len(p->uri.path_raw, start, sep - start);
+			buffer_copy_string_len(p->uri.path, start, sep - start);
 		}
 
 		if (!buffer_is_equal(p->uri.authority, con->uri.authority)) {
@@ -2003,9 +1992,18 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 			return HANDLER_FINISHED;
 		}
 
-		buffer_copy_buffer(p->tmp_buf, p->uri.path_raw);
-		buffer_urldecode_path(p->tmp_buf);
-		buffer_path_simplify(p->uri.path, p->tmp_buf);
+		buffer_urldecode_path(p->uri.path);
+		if (!buffer_is_valid_UTF8(p->uri.path)) {
+			/* invalid UTF-8 after url-decode */
+			con->http_status = 400;
+			return HANDLER_FINISHED;
+		}
+		buffer_path_simplify(p->uri.path, p->uri.path);
+
+		if (buffer_string_is_empty(p->uri.path) || p->uri.path->ptr[0] != '/') {
+			con->http_status = 400;
+			return HANDLER_FINISHED;
+		}
 
 		/* we now have a URI which is clean. transform it into a physical path */
 		buffer_copy_buffer(p->physical.doc_root, con->physical.doc_root);
@@ -2063,13 +2061,7 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 				buffer_copy_buffer(p->physical.path, p->physical.doc_root);
 				buffer_append_slash(p->physical.path);
 				buffer_copy_buffer(p->physical.basedir, p->physical.path);
-
-				/* don't add a second / */
-				if (p->physical.rel_path->ptr[0] == '/') {
-					buffer_append_string_len(p->physical.path, p->physical.rel_path->ptr + 1, buffer_string_length(p->physical.rel_path) - 1);
-				} else {
-					buffer_append_string_buffer(p->physical.path, p->physical.rel_path);
-				}
+				buffer_append_path_len(p->physical.path, CONST_BUF_LEN(p->physical.rel_path));
 			}
 		}
 
@@ -2219,8 +2211,10 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 		}
 
 		return HANDLER_FINISHED;
-	}
-	case HTTP_METHOD_PROPPATCH:
+}
+
+static handler_t mod_webdav_proppatch(server *srv, connection *con, plugin_data *p, handler_ctx *hctx) {
+		struct stat st;
 		if (p->conf.is_readonly) {
 			con->http_status = 403;
 			return HANDLER_FINISHED;
@@ -2237,7 +2231,11 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 			case ENOENT:
 				con->http_status = 404;
 				break;
+			default:
+				con->http_status = 403;
+				break;
 			}
+			return HANDLER_FINISHED;
 		}
 
 		if (S_ISDIR(st.st_mode) && con->physical.path->ptr[buffer_string_length(con->physical.path)-1] != '/') {
@@ -2383,7 +2381,10 @@ propmatch_cleanup:
 #endif
 		con->http_status = 501;
 		return HANDLER_FINISHED;
-	case HTTP_METHOD_LOCK:
+}
+
+#ifdef USE_LOCKS
+static handler_t mod_webdav_lock(server *srv, connection *con, plugin_data *p, handler_ctx *hctx) {
 		/**
 		 * a mac wants to write
 		 *
@@ -2407,17 +2408,18 @@ propmatch_cleanup:
 		 * </D:lockinfo>\n
 		 */
 
+		int depth = mod_webdav_depth(con);
 		if (depth != 0 && depth != -1) {
 			con->http_status = 400;
 
 			return HANDLER_FINISHED;
 		}
 
-#ifdef USE_LOCKS
 		if (con->request.content_length) {
 			xmlDocPtr xml;
 			buffer *hdr_if = NULL;
 			int created = 0;
+			struct stat st;
 
 			if (con->state == CON_STATE_READ_POST) {
 				handler_t r = connection_handle_read_post_state(srv, con);
@@ -2440,6 +2442,12 @@ propmatch_cleanup:
 						return HANDLER_FINISHED;
 					}
 				}
+				else {
+					log_error_write(srv, __FILE__, __LINE__, "sBss",
+							"stat", con->physical.path, ":", strerror(errno));
+					con->http_status = 403; /* Forbidden */
+					return HANDLER_FINISHED;
+				}
 			} else if (hdr_if == NULL && depth == -1) {
 				/* we don't support Depth: Infinity on directories */
 				if (S_ISDIR(st.st_mode)) {
@@ -2503,7 +2511,7 @@ propmatch_cleanup:
 							sqlite3_reset(stmt);
 
 							sqlite3_bind_text(stmt, 1,
-								CONST_BUF_LEN(p->uri.path),
+								CONST_BUF_LEN(con->uri.path),
 								SQLITE_TRANSIENT);
 
 							/* it is the PK */
@@ -2600,7 +2608,7 @@ propmatch_cleanup:
 				return HANDLER_FINISHED;
 			}
 		} else {
-
+			buffer *b;
 			if (NULL != (b = http_header_request_get(con, HTTP_HEADER_OTHER, CONST_STR_LEN("If")))) {
 				buffer *locktoken = b;
 				sqlite3_stmt *stmt = p->conf.stmt_refresh_lock;
@@ -2637,13 +2645,12 @@ propmatch_cleanup:
 				return HANDLER_FINISHED;
 			}
 		}
-		break;
-#else
-		con->http_status = 501;
-		return HANDLER_FINISHED;
+}
 #endif
-	case HTTP_METHOD_UNLOCK:
+
 #ifdef USE_LOCKS
+static handler_t mod_webdav_unlock(server *srv, connection *con, plugin_data *p) {
+		buffer *b;
 		if (NULL != (b = http_header_request_get(con, HTTP_HEADER_OTHER, CONST_STR_LEN("Lock-Token")))) {
 			buffer *locktoken = b;
 			sqlite3_stmt *stmt = p->conf.stmt_remove_lock;
@@ -2690,17 +2697,46 @@ propmatch_cleanup:
 
 			return HANDLER_FINISHED;
 		}
-		break;
-#else
+}
+#endif
+
+SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
+	plugin_data *p = p_d;
+	handler_ctx *hctx = con->plugin_ctx[p->id];
+
+	if (NULL == hctx) return HANDLER_GO_ON;
+	if (!hctx->conf.enabled) return HANDLER_GO_ON;
+	/* physical path is setup */
+	if (buffer_is_empty(con->physical.path)) return HANDLER_GO_ON;
+
+	switch (con->request.http_method) {
+	case HTTP_METHOD_PROPFIND:
+		return mod_webdav_propfind(srv, con, p, hctx);
+	case HTTP_METHOD_MKCOL:
+		return mod_webdav_mkcol(con, p);
+	case HTTP_METHOD_DELETE:
+		return mod_webdav_delete(srv, con, p, hctx);
+	case HTTP_METHOD_PUT:
+		return mod_webdav_put(srv, con, p, hctx);
+	case HTTP_METHOD_MOVE:
+	case HTTP_METHOD_COPY:
+		return mod_webdav_copymove(srv, con, p, hctx);
+	case HTTP_METHOD_PROPPATCH:
+		return mod_webdav_proppatch(srv, con, p, hctx);
+	#ifdef USE_LOCKS
+	case HTTP_METHOD_LOCK:
+		return mod_webdav_lock(srv, con, p, hctx);
+	case HTTP_METHOD_UNLOCK:
+		return mod_webdav_unlock(srv, con, p);
+	#else
+	case HTTP_METHOD_LOCK:
+	case HTTP_METHOD_UNLOCK:
 		con->http_status = 501;
 		return HANDLER_FINISHED;
-#endif
+	#endif
 	default:
-		break;
+		return HANDLER_GO_ON; /* not found */
 	}
-
-	/* not found */
-	return HANDLER_GO_ON;
 }
 
 
