@@ -274,7 +274,7 @@ SETDEFAULTS_FUNC(mod_wstunnel_set_defaults) {
         if (!buffer_is_empty(s->frame_type)
             && !buffer_is_equal_caseless_string(s->frame_type,
                                                 CONST_STR_LEN("binary"))) {
-            buffer_reset(s->frame_type);
+            buffer_clear(s->frame_type);
         }
 
         if (!array_is_vlist(s->origins)) {
@@ -347,7 +347,7 @@ static handler_t wstunnel_recv_parse(server *srv, connection *con, http_response
         DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "fail to send data to client");
         return HANDLER_ERROR;
     }
-    buffer_string_set_length(b, 0);
+    buffer_clear(b);
     UNUSED(srv);
     UNUSED(con);
     return HANDLER_GO_ON;
@@ -496,7 +496,7 @@ static void wstunnel_backend_error(gw_handler_ctx *gwhctx) {
 
 static void wstunnel_handler_ctx_free(void *gwhctx) {
     handler_ctx *hctx = (handler_ctx *)gwhctx;
-    buffer_free(hctx->frame.payload);
+    chunk_buffer_release(hctx->frame.payload);
 }
 
 static handler_t wstunnel_handler_setup (server *srv, connection *con, plugin_data *p) {
@@ -522,11 +522,11 @@ static handler_t wstunnel_handler_setup (server *srv, connection *con, plugin_da
     hctx->gw.create_env       = wstunnel_create_env;
     hctx->gw.handler_ctx_free = wstunnel_handler_ctx_free;
     hctx->gw.backend_error    = wstunnel_backend_error;
-    hctx->gw.response         = buffer_init();
+    hctx->gw.response         = chunk_buffer_acquire();
 
     hctx->frame.state         = MOD_WEBSOCKET_FRAME_STATE_INIT;
     hctx->frame.ctl.siz       = 0;
-    hctx->frame.payload       = buffer_init();
+    hctx->frame.payload       = chunk_buffer_acquire();
 
     binary = !buffer_is_empty(hctx->conf.frame_type); /*("binary")*/
     if (!binary) {
@@ -707,7 +707,7 @@ static int get_key_number(uint32_t *ret, const buffer *b) {
     uint32_t sp = 0;
     char tmp[10 + 1]; /* #define UINT32_MAX_STRLEN 10 */
 
-    for (size_t i = 0, used = b->used; i < used; ++i) {
+    for (size_t i = 0, used = buffer_string_length(b); i < used; ++i) {
         if (light_isdigit(s[i])) {
             tmp[j] = s[i];
             if (++j >= sizeof(tmp)) return -1;
@@ -844,7 +844,7 @@ static int create_response_rfc_6455(handler_ctx *hctx) {
   #endif
 
     value = hctx->srv->tmp_buf;
-    buffer_string_set_length(value, 0);
+    buffer_clear(value);
     buffer_append_base64_encode(value, sha_digest, SHA_DIGEST_LENGTH, BASE64_STANDARD);
     http_header_response_set(con, HTTP_HEADER_OTHER,
                              CONST_STR_LEN("Sec-WebSocket-Accept"),
@@ -938,17 +938,11 @@ static int send_ietf_00(handler_ctx *hctx, mod_wstunnel_frame_type_t type, const
         http_chunk_append_mem(srv, con, &head, 1);
         len = 4*(siz/3)+4+1;
         /* avoid accumulating too much data in memory; send to tmpfile */
-      #if 0
-        chunkqueue_get_memory(con->write_queue, &mem, &len, len, 0);
-        len=li_to_base64(mem,len,(unsigned char *)payload,siz,BASE64_STANDARD);
-        chunkqueue_use_memory(con->write_queue, len);
-      #else
         mem = malloc(len);
         force_assert(mem);
         len=li_to_base64(mem,len,(unsigned char *)payload,siz,BASE64_STANDARD);
         http_chunk_append_mem(srv, con, mem, len);
         free(mem);
-      #endif
         http_chunk_append_mem(srv, con, &tail, 1);
         len += 2;
         break;
@@ -1029,14 +1023,9 @@ static int recv_ietf_00(handler_ctx *hctx) {
                 i++;
                 if (hctx->frame.type == MOD_WEBSOCKET_FRAME_TYPE_TEXT
                     && !buffer_is_empty(payload)) {
-                    size_t len;
                     hctx->frame.ctl.siz = 0;
-                    len = buffer_string_length(payload);
-                    chunkqueue_get_memory(hctx->gw.wb, &mem, &len, len, 0);
-                    len = buffer_string_length(payload);
-                    memcpy(mem, payload->ptr, len);
-                    chunkqueue_use_memory(hctx->gw.wb, len);
-                    buffer_reset(payload);
+                    chunkqueue_append_buffer(hctx->gw.wb, payload);
+                    buffer_clear(payload);
                 }
                 else {
                     if (hctx->frame.state == MOD_WEBSOCKET_FRAME_STATE_INIT
@@ -1044,7 +1033,7 @@ static int recv_ietf_00(handler_ctx *hctx) {
                         buffer *b;
                         size_t len = buffer_string_length(payload);
                         len = (len+3)/4*3+1;
-                        chunkqueue_get_memory(hctx->gw.wb, &mem, &len, len, 0);
+                        chunkqueue_get_memory(hctx->gw.wb, &len);
                         b = hctx->gw.wb->last->mem;
                         len = buffer_string_length(b);
                         DEBUG_LOG(MOD_WEBSOCKET_LOG_DEBUG, "ss",
@@ -1054,7 +1043,7 @@ static int recv_ietf_00(handler_ctx *hctx) {
                                       "fail to base64-decode");
                             return -1;
                         }
-                        buffer_reset(payload);
+                        buffer_clear(payload);
                         /*chunkqueue_use_memory()*/
                         hctx->gw.wb->bytes_in += buffer_string_length(b)-len;
                     }
@@ -1313,15 +1302,9 @@ static int recv_rfc_6455(handler_ctx *hctx) {
                 case MOD_WEBSOCKET_FRAME_TYPE_TEXT:
                 case MOD_WEBSOCKET_FRAME_TYPE_BIN:
                   {
-                    char *mem;
-                    size_t len;
                     unmask_payload(hctx);
-                    len = buffer_string_length(payload);
-                    chunkqueue_get_memory(hctx->gw.wb, &mem, &len, len, 0);
-                    len = buffer_string_length(payload);
-                    memcpy(mem, payload->ptr, len);
-                    chunkqueue_use_memory(hctx->gw.wb, len);
-                    buffer_reset(payload);
+                    chunkqueue_append_buffer(hctx->gw.wb, payload);
+                    buffer_clear(payload);
                     break;
                   }
                 case MOD_WEBSOCKET_FRAME_TYPE_PING:
@@ -1330,11 +1313,11 @@ static int recv_rfc_6455(handler_ctx *hctx) {
                         mod_wstunnel_frame_send(hctx,
                           MOD_WEBSOCKET_FRAME_TYPE_PONG,
                           payload->ptr, buffer_string_length(payload));
-                        buffer_reset(payload);
+                        buffer_clear(payload);
                     }
                     break;
                 case MOD_WEBSOCKET_FRAME_TYPE_PONG:
-                    buffer_reset(payload);
+                    buffer_clear(payload);
                     break;
                 case MOD_WEBSOCKET_FRAME_TYPE_CLOSE:
                 default:
