@@ -24,7 +24,7 @@
 #include <time.h>
 
 int http_response_write_header(server *srv, connection *con) {
-	buffer * const b = buffer_init();
+	buffer * const b = chunkqueue_prepend_buffer_open(con->write_queue);
 
 	if (con->request.http_version == HTTP_VERSION_1_1) {
 		buffer_copy_string_len(b, CONST_STR_LEN("HTTP/1.1 "));
@@ -49,8 +49,7 @@ int http_response_write_header(server *srv, connection *con) {
 	}
 
 	if (304 == con->http_status && (con->response.htags & HTTP_HEADER_CONTENT_ENCODING)) {
-		buffer *vb = http_header_response_get(con, HTTP_HEADER_CONTENT_ENCODING, CONST_STR_LEN("Content-Encoding"));
-		if (NULL != vb) buffer_reset(vb);
+		http_header_response_unset(con, HTTP_HEADER_CONTENT_ENCODING, CONST_STR_LEN("Content-Encoding"));
 	}
 
 	/* add all headers */
@@ -87,10 +86,8 @@ int http_response_write_header(server *srv, connection *con) {
 
 		/* cache the generated timestamp */
 		if (srv->cur_ts != srv->last_generated_date_ts) {
-			buffer_string_prepare_copy(srv->ts_date_str, 255);
-
+			buffer_clear(srv->ts_date_str);
 			buffer_append_strftime(srv->ts_date_str, "%a, %d %b %Y %H:%M:%S GMT", gmtime(&(srv->cur_ts)));
-
 			srv->last_generated_date_ts = srv->cur_ts;
 		}
 
@@ -112,9 +109,7 @@ int http_response_write_header(server *srv, connection *con) {
 		log_error_write(srv, __FILE__, __LINE__, "sSb", "Response-Header:", "\n", b);
 	}
 
-	chunkqueue_prepend_buffer(con->write_queue, b);
-	buffer_free(b);
-
+	chunkqueue_prepend_buffer_commit(con->write_queue);
 	return 0;
 }
 
@@ -401,6 +396,14 @@ handler_t http_response_prepare(server *srv, connection *con) {
 			buffer_copy_buffer(con->uri.path, con->uri.path_raw);
 			buffer_urldecode_path(con->uri.path);
 			buffer_path_simplify(con->uri.path, con->uri.path);
+			if (buffer_string_is_empty(con->uri.path) || con->uri.path->ptr[0] != '/') {
+				log_error_write(srv, __FILE__, __LINE__, "sbs",
+						"uri-path does not begin with '/':", con->uri.path, "-> 400");
+				con->keep_alive = 0;
+				con->http_status = 400;
+				con->file_finished = 1;
+				return HANDLER_FINISHED;
+			}
 		}
 
 		con->conditional_is_valid[COMP_SERVER_SOCKET] = 1;       /* SERVERsocket */
@@ -488,7 +491,7 @@ handler_t http_response_prepare(server *srv, connection *con) {
 		}
 
 		if (con->request.http_method == HTTP_METHOD_OPTIONS &&
-		    con->uri.path->ptr[0] == '*' && con->uri.path_raw->ptr[1] == '\0') {
+		    con->uri.path->ptr[0] == '*' && con->uri.path->ptr[1] == '\0') {
 			/* option requests are handled directly without checking of the path */
 
 			http_header_response_append(con, HTTP_HEADER_OTHER, CONST_STR_LEN("Allow"), CONST_STR_LEN("OPTIONS, GET, HEAD, POST"));
@@ -612,17 +615,7 @@ handler_t http_response_prepare(server *srv, connection *con) {
 
 		buffer_copy_buffer(con->physical.basedir, con->physical.doc_root);
 		buffer_copy_buffer(con->physical.path, con->physical.doc_root);
-		buffer_append_slash(con->physical.path);
-		if (!buffer_string_is_empty(con->physical.rel_path) &&
-		    con->physical.rel_path->ptr[0] == '/') {
-		      #ifdef __COVERITY__
-			if (buffer_string_length(con->physical.rel_path) < 1) return HANDLER_ERROR;
-		      #endif
-			/* coverity[overflow_sink : FALSE] */
-			buffer_append_string_len(con->physical.path, con->physical.rel_path->ptr + 1, buffer_string_length(con->physical.rel_path) - 1);
-		} else {
-			buffer_append_string_buffer(con->physical.path, con->physical.rel_path);
-		}
+		buffer_append_path_len(con->physical.path, CONST_BUF_LEN(con->physical.rel_path));
 
 		if (con->conf.log_request_handling) {
 			log_error_write(srv, __FILE__, __LINE__,  "s",  "-- after doc_root");
