@@ -300,6 +300,7 @@ static server *server_init(void) {
 	srv->srvconf.loadavg[0] = 0.0;
 	srv->srvconf.loadavg[1] = 0.0;
 	srv->srvconf.loadavg[2] = 0.0;
+	srv->srvconf.compat_module_load = 1;
 
 	/* use syslog */
 	srv->errorlog_fd = STDERR_FILENO;
@@ -398,6 +399,7 @@ static void server_free(server *srv) {
 	array_free(srv->split_vals);
 
 	li_rand_cleanup();
+	chunkqueue_chunk_pool_free();
 
 	free(srv);
 }
@@ -962,7 +964,7 @@ static void server_graceful_state (server *srv) {
     else {
         server_sockets_close(srv);
         remove_pid_file(srv);
-        buffer_reset(srv->srvconf.pid_file); /*(prevent more removal attempts)*/
+        buffer_clear(srv->srvconf.pid_file); /*(prevent more removal attempts)*/
     }
 }
 
@@ -1047,6 +1049,19 @@ static int server_main (server * const srv, int argc, char **argv) {
 		}
 	}
 
+      #ifdef __CYGWIN__
+	if (!srv->config_storage && NULL != getenv("NSSM_SERVICE_NAME")) {
+		char *dir = getenv("NSSM_SERVICE_DIR");
+		if (NULL != dir && 0 != chdir(dir)) {
+			log_error_write(srv, __FILE__, __LINE__, "sss", "chdir failed:", dir, strerror(errno));
+			return -1;
+		}
+		srv->srvconf.dont_daemonize = 1;
+		buffer_copy_string_len(srv->srvconf.modules_dir, CONST_STR_LEN("modules"));
+		if (config_read(srv, "conf/lighttpd.conf")) return -1;
+	}
+      #endif
+
 	if (!srv->config_storage) {
 		log_error_write(srv, __FILE__, __LINE__, "s",
 				"No configuration available. Try using -f option.");
@@ -1065,7 +1080,7 @@ static int server_main (server * const srv, int argc, char **argv) {
 	}
 
 	if (test_config) {
-		buffer_reset(srv->srvconf.pid_file);
+		buffer_clear(srv->srvconf.pid_file);
 		if (1 == test_config) {
 			printf("Syntax OK\n");
 		} else { /*(test_config > 1)*/
@@ -1088,7 +1103,7 @@ static int server_main (server * const srv, int argc, char **argv) {
 		graceful_shutdown = 1;
 		srv->sockets_disabled = 1;
 		srv->srvconf.dont_daemonize = 1;
-		buffer_reset(srv->srvconf.pid_file);
+		buffer_clear(srv->srvconf.pid_file);
 		if (srv->srvconf.max_worker) {
 			srv->srvconf.max_worker = 0;
 			log_error_write(srv, __FILE__, __LINE__, "s",
@@ -1163,12 +1178,18 @@ static int server_main (server * const srv, int argc, char **argv) {
 	}
 
 	/* mod_indexfile should be listed in server.modules prior to dynamic handlers */
-	for (i = 0; i < srv->plugins.used; ++i) {
+	i = 0;
+	for (buffer *pname = NULL; i < srv->plugins.used; ++i) {
 		plugin *p = ((plugin **)srv->plugins.ptr)[i];
-		if (buffer_is_equal_string(p->name, CONST_STR_LEN("indexfile"))) break;
+		if (buffer_is_equal_string(p->name, CONST_STR_LEN("indexfile"))) {
+			if (pname) {
+				log_error_write(srv, __FILE__, __LINE__, "SB",
+						"Warning: mod_indexfile should be listed in server.modules prior to mod_", pname);
+			}
+			break;
+		}
 		if (p->handle_subrequest_start && p->handle_subrequest) {
-			log_error_write(srv, __FILE__, __LINE__, "SB",
-					"Warning: mod_indexfile should be listed in server.modules prior to mod_", p->name);
+			if (!pname) pname = p->name;
 		}
 	}
 
@@ -1629,7 +1650,7 @@ static int server_main (server * const srv, int argc, char **argv) {
 			close(pid_fd);
 			pid_fd = -1;
 		}
-		buffer_reset(srv->srvconf.pid_file);
+		buffer_clear(srv->srvconf.pid_file);
 
 		fdevent_clr_logger_pipe_pids();
 		srv->pid = getpid();
@@ -1792,6 +1813,8 @@ static int server_main (server * const srv, int argc, char **argv) {
 				}
 			      #endif
 
+				/* free excess chunkqueue buffers every 64 seconds */
+				if (0 == (min_ts & 0x3f)) chunkqueue_chunk_pool_clear();
 				/* cleanup stat-cache */
 				stat_cache_trigger_cleanup(srv);
 				/* reset global/aggregate rate limit counters */
