@@ -182,7 +182,7 @@ static handler_t stat_cache_handle_fdevent(server *srv, void *_fce, int revent)
 
 				for (j = 0; j < 2; j++) {
 					buffer_copy_string(scf->hash_key, fe.filename);
-					buffer_append_int(scf->hash_key, j);
+					buffer_append_string_len(scf->hash_key, (0 == j ? "0" : "1"), 1);
 
 					ndx = hashme(scf->hash_key);
 
@@ -293,7 +293,7 @@ static handler_t stat_cache_fam_dir_check(server *srv, stat_cache_fam *scf, stat
 	}
 
 	buffer_copy_buffer(scf->hash_key, scf->dir_name);
-	buffer_append_int(scf->hash_key, (int)follow_symlink);
+	buffer_append_string_len(scf->hash_key, (0 != follow_symlink ? "1" : "0"), 1);
 
 	scf->dir_ndx = hashme(scf->hash_key);
 
@@ -546,7 +546,7 @@ const buffer * stat_cache_content_type_get(server *srv, connection *con, const b
 
     if (S_ISREG(sce->st.st_mode)) {
         /* determine mimetype */
-        buffer_reset(sce->content_type);
+        buffer_clear(sce->content_type);
       #if defined(HAVE_XATTR) || defined(HAVE_EXTATTR)
         if (con->conf.use_xattr) {
             stat_cache_attr_get(sce->content_type, name->ptr, srv->srvconf.xattr_name->ptr);
@@ -621,7 +621,7 @@ handler_t stat_cache_get_entry(server *srv, connection *con, buffer *name, stat_
 	sc = srv->stat_cache;
 
 	buffer_copy_buffer(sc->hash_key, name);
-	buffer_append_int(sc->hash_key, follow_symlink);
+	buffer_append_string_len(sc->hash_key, (0 != follow_symlink ? "1" : "0"), 1);
 
 	file_ndx = hashme(sc->hash_key);
 	sc->files = splaytree_splay(sc->files, file_ndx);
@@ -709,9 +709,9 @@ handler_t stat_cache_get_entry(server *srv, connection *con, buffer *name, stat_
 
 	} else {
 
-		buffer_reset(sce->etag);
+		buffer_clear(sce->etag);
 	      #if defined(HAVE_XATTR) || defined(HAVE_EXTATTR)
-		buffer_reset(sce->content_type);
+		buffer_clear(sce->content_type);
 	      #endif
 
 	}
