@@ -1,5 +1,6 @@
 #include "request.h"
 
+#undef NDEBUG
 #include <assert.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -13,9 +14,7 @@ static void test_request_connection_reset(connection *con)
     con->request.http_method = HTTP_METHOD_UNSET;
     con->request.http_version = HTTP_VERSION_UNSET;
     con->request.http_host = NULL;
-    con->request.http_content_type = NULL;
-    con->request.http_if_modified_since = NULL;
-    con->request.http_if_none_match = NULL;
+    con->request.htags = 0;
     con->request.content_length = 0;
     con->header_len = 0;
     con->http_status = 0;
@@ -229,7 +228,7 @@ static void test_request_http_request_parse(server *srv, connection *con)
       array_get_element_klen(con->request.headers, CONST_STR_LEN("Location"));
     assert(ds
            && buffer_is_equal_string(ds->value,
-                                     CONST_STR_LEN("foo, foobar  baz")));
+                                     CONST_STR_LEN("foo, foobar    baz")));
 
     run_http_request_parse(srv, con, __LINE__, 0,
       "#1232 - duplicate headers with line-wrapping - test 2",
@@ -241,7 +240,7 @@ static void test_request_http_request_parse(server *srv, connection *con)
     ds = (data_string *)
       array_get_element_klen(con->request.headers, CONST_STR_LEN("Location"));
     assert(ds
-           && buffer_is_equal_string(ds->value, CONST_STR_LEN("foobar  baz")));
+           && buffer_is_equal_string(ds->value, CONST_STR_LEN("foobar    baz")));
 
     run_http_request_parse(srv, con, __LINE__, 0,
       "#1232 - duplicate headers with line-wrapping - test 3",
@@ -253,7 +252,7 @@ static void test_request_http_request_parse(server *srv, connection *con)
     ds = (data_string *)
       array_get_element_klen(con->request.headers, CONST_STR_LEN("Location"));
     assert(ds
-           && buffer_is_equal_string(ds->value, CONST_STR_LEN("foobar  baz")));
+           && buffer_is_equal_string(ds->value, CONST_STR_LEN("foobar    baz")));
 
     run_http_request_parse(srv, con, __LINE__, 400,
       "missing protocol",
@@ -343,7 +342,7 @@ static void test_request_http_request_parse(server *srv, connection *con)
                     "\r\n"));
     ds = (data_string *)
       array_get_element_klen(con->request.headers, CONST_STR_LEN("ABC"));
-    assert(ds && buffer_is_equal_string(ds->value, CONST_STR_LEN("foo  bc")));
+    assert(ds && buffer_is_equal_string(ds->value, CONST_STR_LEN("foo    bc")));
 
     run_http_request_parse(srv, con, __LINE__, 411,
       "POST request, no Content-Length",
